;**********************************************************************
;                                                                     *
;   This is the code for a voltage booster designed to eliminate      *
;   the voltage drop between a truck and camper/trailer. It also      *
;   functions as a battery charger when plugged into 110vac by        *
;   using the rectified 12V output of the standard power converter    *
;   installed in most campers.                                        *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Filename:	    VBoost.asm                                        *
;    Date:          Sept 19, 2005                                     *
;    File Version:  1.0                                               *
;                                                                     *
;    Author: Kim Christensen                                          *
;    Company:                                                         *
;                                                                     * 
;                                                                     *
;**********************************************************************
;                                                                     *
;    Files required: VBoost.gif (Schematic)                           *
;                    VBoost.txt (Misc. info)                          *
;                                                                     *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Notes:                                                           *
;                                                                     *
;                                                                     *
;                                                                     *
;                                                                     *
;**********************************************************************


	list		p=16f684		; list directive to define processor
	#include	"P16F684.inc"		; processor specific variable definitions
	
	__CONFIG    _CP_OFF & _CPD_OFF & _BOD_ON & _PWRTE_ON & _WDT_OFF & _HS_OSC & _MCLRE_ON & _FCMEN_OFF & _IESO_OFF


; '__CONFIG' directive is used to embed configuration data within .asm file.
; The labels following the directive are located in the respective .inc file.
; See respective data sheet for additional information on configuration word.


#define MAXPWM 0x80
#define MAXCURRENT .150
#define IDLECURRENT .010
#define IDLEPWM .090
#define BULKVOLTAGE .142
#define FLOATVOLTAGE .137
#define VorI 0x0
#define WaitADC 0x1
#define CurrentTimeUp 0x2
#define PluggedInMode 0x3
#define VoltageTimeUp 0x4
#define IdleMode 0x5


;***** VARIABLE DEFINITIONS
w_temp		EQU	0x71			; variable used for context saving
status_temp	EQU	0x72			; variable used for context saving

	cblock 0x20
	Ch04AvgH
	Ch04AvgL
	Ch05AvgH
	Ch05AvgL
	Ch06AvgH
	Ch06AvgL
	AvgCount
	CurrentTLsb
	CurrentTMsb
	VoltageTLsb
	VoltageTMsb
	temp
	temp2
	EvenOdd
	InputVolts
	voltage
	current
	PwmSetting
	TargetVoltage
	modes
	LEDflash
	endc



;**********************************************************************
	ORG		0x000			; processor reset vector
  	goto		main			; go to beginning of program


	ORG		0x004			; interrupt vector location
	movwf		w_temp			; save off current W register contents
	movf		STATUS,w		; move status register into W register
	movwf		status_temp		; save off contents of STATUS register

; isr code can go here or be located as a call subroutine elsewhere
	;(Start ISR  ~ 76.3Hz rate)

	BCF STATUS, RP0 	; Bank0
	decfsz VoltageTLsb, F	;This routine is a divide by 4578.
	goto nextTimer
	movlw .77		
;	movlw .5			;Temporary to speed simulation (4.8 seconds)
	movwf VoltageTLsb
	decfsz VoltageTMsb, F	;This happens apx every second
	goto nextTimer
	bsf modes, VoltageTimeUp

nextTimer
	decfsz CurrentTLsb, F	;This routine is a divide by 4578.
	goto endisr
	movlw .77		
;	movlw .5			;Temporary to speed simulation (4.8 seconds)
	movwf CurrentTLsb
	decfsz CurrentTMsb, F	;This happens apx every second
	goto endisr
	bsf modes, CurrentTimeUp

endisr

	bcf PIR1, TMR1IF		; Clear timer overflow bit
	;(End   ISR)
	;
	movf		status_temp,w		; retrieve copy of STATUS register
	movwf		STATUS			; restore pre-isr STATUS register contents
	swapf		w_temp,f
	swapf		w_temp,w		; restore pre-isr W register contents
	retfie					; return from interrupt


main

; remaining code goes here

	BCF STATUS,RP0 	;Bank 0
	CLRF PORTC 		;Init PORTC
	CLRF PORTA 		;Init PORTA
	MOVLW 07h 		;Set RC<4,1:0> to
	MOVWF CMCON0 	;digital I/O
	BSF STATUS,RP0 	;Bank 1
	MOVLW B'01100000' ;A/D clock speed
	MOVWF ADCON1
	movlw B'01110000' ; analog on RC0, 1 & 2
	movwf ANSEL 	;digital I/O on rest
	movlw B'00010111' ; RC0, 1, 2 & 5 are inputs
	movwf TRISC
	movlw B'00111100' ; RA0 & 1 are outputs
	movwf TRISA
	BCF STATUS,RP0	;Bank 0

	movlw PIE1
	movwf FSR
	bsf INDF, TMR1IF		; Turn on timer 1 interupts
	bsf INTCON, PEIE		; Turn on perifial interupts
	bsf INTCON, GIE			; Turn on global interupts

	movlw 0x01
	movwf T1CON			; Start timer 1

	movlw B'00010001'			; Power up ADC and set to channel 4
	movwf ADCON0


	CLRF CCP1CON 		; CCP Module is off
	CLRF TMR2 			; Clear Timer2
	MOVLW 0x3E 			; 78.2Khz period with 8bit resolution
	BSF STATUS, RP0 	; Bank1
	MOVWF PR2 
	BCF STATUS, RP0 	; Bank0
	MOVLW B'00000000'
	MOVWF CCPR1L 		; Duty Cycle is 50% of PWM Period
	MOVLW B'10001101' 	; PWM mode, 2 LSbs of Duty cycle = 00
	MOVWF CCP1CON 
	movlw 0x7C			; Set timer2 on and post scaler to 1:16
	movwf T2CON			; Timer2 starts to increment

	movlw BULKVOLTAGE
	movwf TargetVoltage
	clrf PwmSetting
	clrf modes			; Voltage mode, Bulk charging,
	call ClrAvgMem
	call ResetCurrTimer
waitonemin
	call adcRoutine
	btfss modes, CurrentTimeUp
	goto waitonemin

	movlw .100
	subwf InputVolts, W
	bc TruckMode
	bsf modes, PluggedInMode	; Set to Plugged in mode if truck pwr less than 10.0V
	movlw B'00100111' 			; RC0, 1, 2 & 6 are inputs
	BSF STATUS,RP0 	;Bank 1
	movwf TRISC
	BCF STATUS,RP0	;Bank 0
TruckMode
	call ResetCurrTimer
	call ResetVoltTimer

delay
	call adcRoutine
	; we are here every 38.7ms

	movlw FLOATVOLTAGE
	subwf TargetVoltage, W
	movlw 0x01
	skpz
	movlw 0x02
	btfss modes, IdleMode
	goto SetLEDs
	incf LEDflash, F
	btfss LEDflash, 0x4
	movlw 0x00
SetLEDs
	movwf PORTA


	btfss modes, PluggedInMode
	goto chkTruckBat
	movlw IDLEPWM
	subwf PwmSetting, W			; Check if PWM (current) has dropped enough to enter float voltage mode
	btfsc STATUS, C
	call ResetCurrTimer			; Reset timer if current goes above IDLEPWM level during 1 minute interval
	movlw FLOATVOLTAGE
	btfsc modes, CurrentTimeUp
	movwf TargetVoltage			; Set to float voltage mode if over 1 minute of IDLEPWM detected
	goto VoltageMode
chkTruckBat
	rrf current, W				; Take current reading and divide by 2
	movwf temp
	rrf temp, F					; and again...
	rrf temp, W					; and again and place into W (current/8 == apx wiring voltage loss)
	andlw B'00011111'			; Mask off any stray carry bits that got rotated in
	addwf InputVolts, W			; Add the Truck --> Camper wiring voltage drop to camper input voltage.
	skpnc						; Do next step if voltage over 25.5V!!!
	movlw .255					; This should never happen, but just in case!
	btfsc modes, IdleMode		; Do Idle routine if truck voltage has gone below 12.8V level for set period
	goto idle
	sublw .128					; check if estimated truck battery voltage is less than 12.8V
;	sublw .100					; check if estimated truck battery voltage is less than 10.0V
	skpc
	call ResetVoltTimer			; Reset timer if truck voltage goes above 12.8V level during 1 minute interval
	btfss modes, VoltageTimeUp	; check if timer expired
	goto chkcurrent
	clrf PwmSetting				; Set drive to zero
	bsf modes, IdleMode			; and enter idle state.
	goto pwm
idle
	sublw .137					; check if estimated truck battery voltage is greater than 13.7V
	skpnc
	call ResetVoltTimer			; Reset timer if truck voltage goes below 13.7V level during 1 minute interval
	btfsc modes, VoltageTimeUp	; check if timer expired & if yes:
	bcf modes, IdleMode			; exit idle state and return to voltage boosting.
	call ResetCurrTimer			; Reset so we don't goto floatmode accidentally
	goto pwm

chkcurrent
	movlw IDLECURRENT
	subwf current, W			; Check if current has dropped enough to enter float voltage mode
	btfsc STATUS, C
	call ResetCurrTimer			; Reset timer if current goes above IDLECURRENT level during 1 minute interval
	movlw FLOATVOLTAGE
	btfsc modes, CurrentTimeUp
	movwf TargetVoltage			; Set to float voltage mode if over 1 minute of IDLECURRENT detected

	movlw MAXCURRENT
	subwf current, W			; Check if current limiting is required
	btfsc STATUS, C
	bsf modes, VorI				; Set to current LTD mode if true
	btfss modes, VorI			; Check if already in current LTD mode.
	goto VoltageMode

	movf TargetVoltage, W
	subwf voltage, W			; voltage - TargetVoltage = error voltage in W
	btfsc STATUS, C
	bcf modes, VorI				; If voltage too high, set to constant voltage mode for next pass	
	movlw MAXCURRENT
	subwf current, W			; Check how much current limiting is required
	btfsc STATUS, Z
	goto delay					; If no current error, then do nothing.
	btfss STATUS, C
	goto up
	goto down	

VoltageMode
	movf TargetVoltage, W
	subwf voltage, W			; voltage - TargetVoltage = error voltage in W
	btfsc STATUS, Z
	goto delay					; If no error voltage, then do nothing.
	btfss STATUS, C
	goto up
down
	
	movwf temp
	bcf STATUS, C
	btfsc modes, VorI			; Check if in current LTD mode.
	rrf temp, F					; If so, divide error current by 2
	bcf STATUS, C
	rrf temp, F					; divide by 2 and leave in "temp" because we must
	movf temp, W				; move it anyway just to effect the damn Z flag
	btfsc STATUS, Z
	movlw 0x01					; Don't let it go to zero or nothing will get done!
	
	subwf PwmSetting, F			; Ramp the PWM setting down in proportion to error voltage/current
	btfss STATUS, C
	clrf PwmSetting
	goto pwm
up
	
	movwf temp
	bsf STATUS, C
	btfsc modes, VorI			; Check if in current LTD mode.
	rrf temp, F					; If so, divide error current by 2
	bsf STATUS, C
	rrf temp, W					; divide by 2 No need to check for zero because 0xFF is -1
	
	subwf PwmSetting, F			; Ramp the PWM setting up in proportion to error voltage/current:
	btfss modes, PluggedInMode	; We have subtracted a negative value. Effectively adding the absolute value.
	goto truckpwmmode
	movlw 0xFF					; Full %100 duty cycle allowed for plugged in mode
	btfsc STATUS, C				; Check if value over flowed.				
	movwf PwmSetting			; and set max pwm value if it did.
	goto pwm
truckpwmmode
	btfsc STATUS, C				; Check if value over flowed.
	goto SetMaxPwm				; and set max pwm value if it did.
	movf PwmSetting, W
	sublw MAXPWM
	btfsc STATUS, C	
	goto pwm
SetMaxPwm
	movlw MAXPWM				; Set the maximun permissable duty cycle
	movwf PwmSetting

pwm
	movf PwmSetting, W
	movwf temp			; make a copy of pwm setting
	bcf STATUS, C
	rrf temp, F
	bcf STATUS, C
	rrf temp, F
;	MOVLW 0x8C 			; PWM mode, 2 LSbs of Duty cycle = 00
	MOVLW B'10001101' 	; PWM mode, 2 LSbs of Duty cycle = 00
	btfsc PwmSetting, 0x0
	iorlw 0x10
	btfsc PwmSetting, 0x1
	iorlw 0x20
	movwf CCP1CON			; Set 2 LSBs of duty cycle
	movf temp, W
	movwf CCPR1L			; Set MSBs of duty cycle
;around
;	BCF PIR1, TMR2IF		; Clear timer2 overflow flag
	goto	delay









adcRoutine
	BTFSS PIR1, TMR2IF		; Check timer2 overflow flag
	goto adcRoutine
	BCF PIR1, TMR2IF		; Clear timer2 overflow flag
	incf EvenOdd, F
	btfsc EvenOdd, 0		; Only take readings every 2nd time.
	goto chkadc
	bsf ADCON0, 1			; Set GO/DONE bit to start ADC conversion sequence
	goto adcRoutine			; and wait for conversion to finish
chkadc
	movf ADRESH, W			; Get ADC reading after selected input has been sampled by ADC....
	btfsc ADCON0, CHS1
	goto chan6
	btfsc ADCON0, CHS0
	goto chan5
chan4

	addwf Ch04AvgL, F			; Channel 4 (InputVolts)
	btfsc STATUS, C
	incf  Ch04AvgH, F
	;movwf InputVolts
	movlw B'00010101'
	goto setchannel
chan5
	addwf Ch05AvgL, F			; Channel 5 (voltage)
	btfsc STATUS, C
	incf  Ch05AvgH, F
	;movwf voltage
	movlw B'00011001'
	goto setchannel
chan6
	addwf Ch06AvgL, F			; Channel 6 (current)
	btfsc STATUS, C
	incf  Ch06AvgH, F
	;movwf current
	decfsz AvgCount, F
	goto  AvgLater

;	movlw .6				; Set the divizor (power of 2) to 64
	movlw .5				; Set the divizor (power of 2) to 32
	movwf AvgCount
CalcAvg
	rrf Ch04AvgH, F			; Divide results by 32 (Number in AvgCount)
	rrf Ch04AvgL, F
	rrf Ch05AvgH, F			; We really don't care about the carry bit from prev result
	rrf Ch05AvgL, F			; because the "Ch0xAvgH" result will be discarded anyway.
	rrf Ch06AvgH, F
	rrf Ch06AvgL, F
	decfsz AvgCount, F
	goto CalcAvg

	movf Ch04AvgL, W
	movwf InputVolts
	movf Ch05AvgL, W
	movwf voltage
	movf Ch06AvgL, W
	movwf current
	call  ClrAvgMem
	movlw B'00010001'
	movwf ADCON0			; Set new channel for next ADC conversion and wait for inputs to settle
	return

AvgLater
	movlw B'00010001'
setchannel
	movwf ADCON0			; Set new channel for next ADC conversion and wait for inputs to settle
	goto adcRoutine

ClrAvgMem
;	movlw .64
	movlw .32
	movwf AvgCount
	clrf Ch04AvgL 
	clrf Ch04AvgH 
	clrf Ch05AvgL 
	clrf Ch05AvgH 
	clrf Ch06AvgL 
	clrf Ch06AvgH 
	return

ResetCurrTimer
	bcf INTCON, GIE			; Turn off global interupts
	bcf modes, CurrentTimeUp
	movlw .77		
	movwf CurrentTLsb
	movlw .60
	movwf CurrentTMsb
	bsf INTCON, GIE			; Turn on global interupts
	return

ResetVoltTimer
	bcf INTCON, GIE			; Turn off global interupts
	bcf modes, VoltageTimeUp
	movlw .77		
	movwf VoltageTLsb
	movlw .60
	movwf VoltageTMsb
	bsf INTCON, GIE			; Turn on global interupts
	return

	;ORG	0x2100				; data EEPROM location
	;DE	1,2,3,4				; define first four EEPROM locations as 1, 2, 3, and 4




	END                       ; directive 'end of program'

